"""
Detect which kind of structure we're dealing with
"""
from soorgeon import get


def is_f_string(leaf):
    return leaf.type == 'fstring_start'


def is_funcdef(leaf):
    """
    Returns true if the leaf is the beginning of a function definition (def
    keyword)
    """
    return leaf.type == 'keyword' and leaf.value == 'def'


def is_lambda(leaf, raise_=False):
    """
    Returns true if the leaf is the beginning of a lambda definition
    """
    return leaf.type == 'keyword' and leaf.value == 'lambda'


def is_classdef(leaf):
    """
    Returns true if the leaf is the beginning of a class definition
    """
    return leaf.type == 'keyword' and leaf.value == 'class'


def is_for_loop(leaf):
    """
    Returns true if the leaf is the beginning of a for loop
    """
    has_suite_parent = False
    parent = leaf.parent

    while parent:
        if parent.type == 'suite':
            has_suite_parent = True

        if parent.type == 'for_stmt':
            return not has_suite_parent

        parent = parent.parent

    return False


def is_comprehension(leaf):
    """
    Return true if the leaf is the beginning of a list/set/dict comprehension.
    Returns true for generators as well
    """
    if leaf.type != 'operator' or leaf.value not in {'[', '(', '{'}:
        return False

    sibling = leaf.get_next_sibling()

    return (sibling.type in {'testlist_comp', 'dictorsetmaker'}
            and sibling.children[-1].type == 'sync_comp_for')


def is_context_manager(leaf):
    """
    Returns true if the leaf is the beginning of a context manager
    """
    has_suite_parent = False
    parent = leaf.parent

    while parent:
        if parent.type == 'suite':
            has_suite_parent = True

        if parent.type == 'with_stmt':
            return not has_suite_parent

        parent = parent.parent

    return False


def is_left_side_of_assignment(node):
    to_check = get.first_expr_stmt_parent(node)

    if not to_check:
        return False

    return to_check.children[1].value == '='


# FIXME: delete
def is_inside_list_comprehension(node):
    parent = get.first_non_atom_expr_parent(node)

    return (parent.type == 'testlist_comp'
            and parent.children[1].type == 'sync_comp_for')


def is_inside_funcdef(leaf):
    parent = leaf.parent

    while parent:
        if parent.type == 'funcdef':
            return True

        parent = parent.parent

    return False


def is_inside_function_call(leaf):
    # ignore it if this is a function definition
    if leaf.parent.type == 'param':
        return False

    next_sibling = leaf.get_next_sibling()

    try:
        next_sibling_value = next_sibling.value
    except AttributeError:
        next_sibling_value = None

    # ignore names in keyword arguments
    # e.g., some_function(x=1)
    # (x does not count since it's)
    if next_sibling_value == '=':
        return False

    # check if the node is inside parenhesis: function(df)
    # or a parent of the node: function(df.something)
    # NOTE: do we need more checks to ensure we're in the second case?
    # maybe check if we have an actual dot, or we're using something like
    # df[key]?

    node = leaf

    while node:
        inside = is_inside_parenthesis(node)

        if inside:
            return True
        else:
            node = node.parent

    return False


def is_inside_parenthesis(node):
    try:
        left = node.get_previous_sibling().value == '('
    except AttributeError:
        left = False

    try:
        right = node.get_next_sibling().value == ')'
    except AttributeError:
        right = False

    try:
        # to prevent (1, 2, 3) being detected as a function call
        has_name = node.get_previous_sibling().get_previous_leaf(
        ).type == 'name'
    except AttributeError:
        has_name = False

    return left and right and has_name


def is_accessing_variable(leaf):
    """
    For a given node of type name, determine if it's used
    """
    # NOTE: what if we only have the name and we are not doing anything?
    # like:
    # df
    # that still counts as dependency
    try:
        children = leaf.get_next_sibling().children
    except Exception:
        return False

    getitem = children[0].value == '[' and children[-1].value == ']'
    dotaccess = children[0].value == '.'
    # FIXME: adding dotacess breaks other tests
    return getitem or dotaccess



import typing as t
from click.exceptions import ClickException
from click._compat import get_text_stderr
from click.utils import echo
from gettext import gettext as _


def _format_message(exception):
    if hasattr(exception, 'format_message'):
        return exception.format_message()
    else:
        return str(exception)


def _build_message(exception):
    msg = _format_message(exception)

    while exception.__cause__:
        msg += f'\n{_format_message(exception.__cause__)}'
        exception = exception.__cause__

    return msg


class BaseException(ClickException):
    """
    A subclass of ClickException that adds support for printing error messages
    from chained exceptions
    """

    def show(self, file: t.Optional[t.IO] = None) -> None:
        if file is None:
            file = get_text_stderr()

        message = _build_message(self)
        echo(_("Error: {message}").format(message=message), file=file)


class InputWontRunError(BaseException):
    """Raised when there are errors that make running the input infeasible
    """
    pass


class InputError(BaseException):
    """Raised when the input has issues and needs user's editing
    """
    pass


class InputSyntaxError(InputWontRunError):
    """Raised if the notebook has invalid syntax
    """
    pass



"""
High-level description of the refactoring process

Given some notebook:

```python
# H1 header

# H2 header
x = 1 + 1

# H2 header
y = x + 1

# H3 header
z = x + 1
```

Step 1. Section splitting.

We split it into sections using H2 markdown headers (H1 goes into the first
task), we call this a proto-task:


```python
# H1 header

# H2 header
x = 1 + 1
```

```python
# H2 header
y = x + 1
```

```python
# H3 header
z = x + 1
```

Step 2. Parse inputs and outputs.

Then we figure out which variables each proto-task uses and which ones they
create:

1. Exposes `x`
2. Exposes `y`, uses `x`
3. Exposes `z`, uses `x`


Step 3. Resolve dependencies.

Then we resolve dependencies:

1 -> 2
1 -> 3


Step 4. Determine variables to serialize.

Now we determine the products for each proto-task, ignoring the variables that
aren't used downstream:

1. x
2. None
3. None


Step 5. Add serialization cell.

Then, we modify each task to serialize the outputs we need:

```python
# added by soorgeon
from pathlib import Path
import pickle

# H1 header

# H2 header
x = 1 + 1

# added by soorgeon
Path(product['x']).write_bytes(x)
```

Step 6. Add parameters cell with upstram dependencies.

We also add the parameters cell with the upstream dependdencies:

```python
# %% tags=["parameters"]
upstream = ['one']

# H2 header
y = x + 1
```

Step 7. Refactor imports.

We also need to modify imports, we parse all imports in the
notebooks, delete them, and then add a new cell at the top with the
subset of import statements that the proto task uses.

```python
# soorgeon auto-generated imports statements
import pandas as pd

# %% tags=["parameters"]
upstream = ['one']

# H2 header
y = x + 1
```

Step 8. Generate step.

Finally, we generate the pipeline.yaml file.
"""
import shutil
import traceback
import ast
import pprint
from collections import namedtuple
from pathlib import Path
import logging
from importlib import resources
from soorgeon import assets

import click
import parso
import jupytext
import yaml
import nbformat

from soorgeon import (split, io, definitions, proto, exceptions, magics,
                      pyflakes)

logger = logging.getLogger(__name__)
pp = pprint.PrettyPrinter(indent=4)


class NotebookExporter:
    """Converts a notebook into a Ploomber pipeline
    """
    def __init__(self, nb, verbose=True, df_format=None, py=False):
        if df_format not in {None, 'parquet', 'csv'}:
            raise ValueError("df_format must be one of "
                             "None, 'parquet' or 'csv', "
                             f"got: {df_format!r}")

        # NOTE: we're commenting magics here but removing them in ProtoTask,
        # maybe we should comment magics also in ProtoTask?
        nb = magics.comment_magics(nb)

        self._nb = nb
        self._df_format = df_format
        self._verbose = verbose

        self._io = None
        self._definitions = None
        self._tree = None
        self._providers = None

        self._check()

        self._proto_tasks = self._init_proto_tasks(nb, py)

        # snippets map names with the code the task will contain, we use
        # them to run static analysis
        self._snippets = {pt.name: str(pt) for pt in self._proto_tasks}

    def export(self, product_prefix=None):
        """Export the project

        Parameters
        ---------
        product_prefix : str
            A prefix to append to all products. If None, it is set to 'output'
        """
        product_prefix = product_prefix or 'output'

        # export functions and classes to a separate file
        self.export_definitions()

        # export requirements.txt
        self.export_requirements()

        # export .gitignore
        self.export_gitignore(product_prefix)

        task_specs = self.get_task_specs(product_prefix=product_prefix)

        sources = self.get_sources()

        dag_spec = {'tasks': list(task_specs.values())}

        for name, task_spec in task_specs.items():
            path = Path(task_spec['source'])
            path.parent.mkdir(exist_ok=True, parents=True)
            path.write_text(sources[name])

        out = yaml.dump(dag_spec, sort_keys=False)
        # pyyaml doesn't have an easy way to control whitespace, but we want
        # tasks to have an empty line between them
        out = out.replace('\n- ', '\n\n- ')

        Path('pipeline.yaml').write_text(out)

        self.export_readme()

    def _check(self):
        """
        Run a few checks before continuing the refactoring. If this fails,
        we'll require the user to do some small changes to their code.
        """
        code = self._get_code()
        _check_syntax(code)
        pyflakes.check_notebook(self._nb)
        _check_functions_do_not_use_global_variables(code)
        _check_no_star_imports(code)

    def _init_proto_tasks(self, nb, py):
        """Break notebook into smaller sections
        """
        # use H2 headers to break notebook
        breaks = split.find_breaks(nb)

        # generate groups of cells
        cells_split = split.split_with_breaks(nb.cells, breaks)

        # extract names by using the H2 header text
        names = split.names_with_breaks(nb.cells, breaks)

        # initialize proto tasks
        return [
            proto.ProtoTask(
                name,
                cell_group,
                df_format=self._df_format,
                py=py,
            ) for name, cell_group in zip(names, cells_split)
        ]

    def get_task_specs(self, product_prefix=None):
        """Return task specs (dictionary) for each proto task
        """
        return {
            pt.name: pt.to_spec(self.io, product_prefix=product_prefix)
            for pt in self._proto_tasks
        }

    def get_sources(self):
        """
        Generate the code strings (ipynb or percent format) for each proto task
        """
        # FIXME: this calls find_providers, we should only call it once
        upstream = io.find_upstream(self._snippets)

        code_nb = self._get_code()

        return {
            pt.name: pt.export(
                upstream,
                self.io,
                self.providers,
                code_nb,
                self.definitions,
            )
            for pt in self._proto_tasks
        }

    def export_definitions(self):
        """Create an exported.py file with function and class definitions
        """
        # do not create exported.py if there are no definitions
        if not self.definitions:
            return

        out = '\n\n'.join(self.definitions.values())

        ip = io.ImportsParser(self._get_code())
        imports = ip.get_imports_cell_for_task(out)

        if imports:
            exported = f'{imports}\n\n\n{out}'
        else:
            exported = out

        Path('exported.py').write_text(exported)

    def export_requirements(self):
        """Generates requirements.txt file, appends it at the end if already
        exists
        """
        reqs = Path('requirements.txt')

        # ploomber is added by default (pinned to >=0.14.7 because earlier
        # versions throw an error when using the inline bash IPython magic
        # during the static_analysis stage)
        pkgs = ['ploomber>=0.14.7'] + definitions.packages_used(self.tree)

        # add pyarrow to requirements if needed
        if (self._df_format == 'parquet' and 'pyarrow' not in pkgs
                and 'fastparquet' not in pkgs):
            pkgs = ['pyarrow'] + pkgs

        pkgs_txt = '\n'.join(sorted(pkgs))

        out = f"""\
# Auto-generated file, may need manual editing
{pkgs_txt}
"""
        if reqs.exists():
            reqs.write_text(reqs.read_text() + out)
        else:
            reqs.write_text(out)

    def _get_code(self):
        """Returns the source of code cells
        """
        return '\n'.join(cell['source'] for cell in self._nb.cells
                         if cell['cell_type'] == 'code')

    def export_gitignore(self, product_prefix):
        if product_prefix and not Path(product_prefix).is_absolute():
            path = Path('.gitignore')
            content = '' if not path.exists() else path.read_text() + '\n'
            path.write_text(content + product_prefix + '\n')
            self._echo(f'Added {str(product_prefix)!r} directory'
                       ' to .gitignore...')

    def export_readme(self):
        path = Path('README.md')

        if path.exists():
            content = path.read_text() + '\n'
            self._echo('README.md found, appended auto-generated content')
        else:
            content = ''
            self._echo('Added README.md')

        path.write_text(content + resources.read_text(assets, 'README.md'))

    def _echo(self, msg):
        if self._verbose:
            click.echo(msg)

    @property
    def definitions(self):
        if self._definitions is None:
            self._definitions = (definitions.from_def_and_class(self.tree))

        return self._definitions

    @property
    def tree(self):
        if self._tree is None:
            code = self._get_code()
            self._tree = parso.parse(code)

        return self._tree

    @property
    def providers(self):
        if self._providers is None:
            self._providers = io.ProviderMapping(self.io)

        return self._providers

    @property
    def io(self):
        """
        {name: (inputs, outputs), ...}
        """
        if self._io is None:
            io_ = self._get_raw_io()

            logging.info(f'io: {pp.pformat(io_)}\n')

            self._io = io.prune_io(io_)

            logging.info(f'pruned io: {pp.pformat(self._io)}\n')

        return self._io

    def _get_raw_io(self):
        return io.find_io(self._snippets)


FunctionNeedsFix = namedtuple('FunctionNeedsFix', ['name', 'pos', 'args'])


def _check_syntax(code):
    try:
        ast.parse(code)
    except SyntaxError:
        error = traceback.format_exc()
    else:
        error = None

    if error:
        raise exceptions.InputSyntaxError(f'Could not refactor notebook due '
                                          f'to invalid syntax\n\n {error}')


def _check_no_star_imports(code):
    tree = parso.parse(code)

    star_imports = [
        import_ for import_ in tree.iter_imports() if import_.is_star_import()
    ]

    if star_imports:
        star_imports_ = '\n'.join(import_.get_code()
                                  for import_ in star_imports)
        url = ('https://github.com/ploomber/soorgeon/blob/main/doc'
               '/star-imports.md')
        raise exceptions.InputError(
            'Star imports are not supported, please change '
            f'the following:\n\n{star_imports_}\n\n'
            f'For more details, see: {url}')


# see issue #12 on github
def _check_functions_do_not_use_global_variables(code):
    tree = parso.parse(code)

    needs_fix = []

    local_scope = set(definitions.find_defined_names(tree))

    for funcdef in tree.iter_funcdefs():
        # FIXME: this should be passing the tree directly, no need to reparse
        # again, but for some reason,
        # using find_inputs_and_outputs_from_tree(funcdef) returns the name
        # of the function as an input
        in_, _ = io.find_inputs_and_outputs(funcdef.get_code(),
                                            local_scope=local_scope)

        if in_:
            needs_fix.append(
                FunctionNeedsFix(
                    funcdef.name.value,
                    funcdef.start_pos,
                    in_,
                ))

    if needs_fix:
        message = ('Looks like the following functions are using global '
                   'variables, this is unsupported. Please add all missing '
                   'arguments. See this to learn more:\n'
                   'https://github.com/ploomber/soorgeon/blob'
                   '/main/doc/fn-global.md\n\n')

        def comma_separated(args):
            return ','.join(f"'{arg}'" for arg in args)

        message += '\n'.join(
            f'* Function {f.name!r} uses variables {comma_separated(f.args)}'
            for f in needs_fix)

        raise exceptions.InputError(message)


def from_nb(nb, log=None, product_prefix=None, df_format=None, py=False):
    """Refactor a notebook by passing a notebook object

    Parameters
    ----------
    product_prefix : str
        A prefix to add to all products. If None, it's set to 'output'

    """
    if log:
        logging.basicConfig(level=log.upper())

    exporter = NotebookExporter(nb, df_format=df_format, py=py)

    exporter.export(product_prefix=product_prefix)

    # TODO: instantiate dag since this may raise issues and we want to capture
    # them to let the user know how to fix them (e.g., more >1 H2 headers with
    # the same text)


def from_path(path, log=None, product_prefix=None, df_format=None, py=False):
    """Refactor a notebook by passing a path to it

    Parameters
    ----------
    allow_single_task : bool
        If False, the function will fail if it cannot refactor the notebook
        into a multi-stage pipeline. If True, it will first try to refactor
        the notebook, and if it fails, it will generate a pipeline with
        a single task
    """
    from_nb(jupytext.read(path),
            log=log,
            product_prefix=product_prefix,
            df_format=df_format,
            py=py)


def single_task_from_path(path, product_prefix, file_format):
    """Refactor a notebook into a single task Ploomber pipeline
    """
    path = Path(path)

    click.echo('Creating a pipeline with a single task...')

    cell = nbformat.v4.new_code_cell(source='upstream = None',
                                     metadata=dict(tags=['parameters']))

    nb = jupytext.read(path)
    nb.cells.insert(0, cell)

    name = path.stem
    path_backup = path.with_name(f'{name}-backup{path.suffix}')

    # output
    ext = path.suffix[1:] if file_format is None else file_format
    path_to_task = f'{name}.{ext}'

    # create backup
    shutil.copy(path, path_backup)

    jupytext.write(nb,
                   path_to_task,
                   fmt='py:percent' if ext == 'py' else 'ipynb')

    spec = {
        'tasks': [{
            'source':
            path_to_task,
            'product':
            str(Path(product_prefix or 'products', f'{name}-report.ipynb'))
        }]
    }

    pipeline = 'pipeline.yaml'
    click.echo(f'Done. Copied code to {path_to_task!r} and added it to '
               f'{pipeline!r}. Created backup of original notebook '
               f'at {str(path_backup)!r}.')

    Path('pipeline.yaml').write_text(yaml.safe_dump(spec, sort_keys=False))


def refactor(path, log, product_prefix, df_format, single_task, file_format):

    if single_task:
        single_task_from_path(path=path,
                              product_prefix=product_prefix,
                              file_format=file_format)
    else:
        ext = Path(path).suffix[1:] if file_format is None else file_format

        try:
            from_nb(jupytext.read(path),
                    log=log,
                    product_prefix=product_prefix,
                    df_format=df_format,
                    py=ext == 'py')
        # InputError means the input is broken
        except exceptions.InputWontRunError:
            raise
        # This implies an error on our end
        except Exception as e:
            logger.exception('Error calling from_nb')
            cmd = f'soorgeon refactor {path} --single-task'
            msg = ('An error occurred when refactoring '
                   'notebook.\n\nTry refactoring '
                   f'as a single task pipeline:\n\n$ {cmd}\n\n'
                   'Need help? https://ploomber.io/community\n\n'
                   'Error details:\n')
            raise exceptions.InputError(msg) from e



import click

from soorgeon import __version__, export


@click.group()
@click.version_option(__version__)
def cli():
    pass


@cli.command()
@click.argument('path', type=click.Path(exists=True))
@click.option('--log', '-l', default=None)
@click.option(
    '--df-format',
    '-d',
    default=None,
    type=click.Choice(('parquet', 'csv')),
    help='Format for variables with the df prefix. Otherwise uses pickle')
@click.option('--product-prefix',
              '-p',
              default=None,
              help='Prefix for all products')
@click.option('--single-task',
              '-s',
              is_flag=True,
              help='Create a pipeline with a single task')
@click.option(
    '--file-format',
    '-f',
    default=None,
    type=click.Choice(('py', 'ipynb')),
    help=('Format for pipeline tasks, if empty keeps the same format '
          'as the input'))
def refactor(path, log, product_prefix, df_format, single_task, file_format):
    """
    Refactor a monolithic notebook.

    $ soorgeon refactor nb.ipynb

    * Sections must be separated by markdown H2 headings

    * Star imports (from math import *) not supported

    * Functions should not use global variables

    User guide: https://github.com/ploomber/soorgeon/blob/main/doc/guide.md
    """
    export.refactor(path,
                    log,
                    product_prefix=product_prefix,
                    df_format=df_format,
                    single_task=single_task,
                    file_format=file_format)

    click.secho(f'Finished refactoring {path!r}, use Ploomber to continue.',
                fg='green')

    click.echo("""
Install dependencies (this will install ploomber):
    $ pip install -r requirements.txt

List tasks:
    $ ploomber status

Execute pipeline:
    $ ploomber build

Plot pipeline (this requires pygraphviz, which isn't installed by default):
    $ ploomber plot

* Documentation: https://docs.ploomber.io
* Jupyter integration: https://ploomber.io/s/jupyter
* Other editors: https://ploomber.io/s/editors
""")



import copy
import re

_IS_IPYTHON_CELL_MAGIC = r'^\s*%{2}[a-zA-Z]+'
_IS_IPYTHON_LINE_MAGIC = r'^\s*%{1}[a-zA-Z]+'
_IS_INLINE_SHELL = r'^\s*!{1}.+'

_IS_COMMENTED_LINE_MAGIC = r'^(.+) # \[magic\] (%.*)'

_PREFIX = '# [magic] '
_PREFIX_LEN = len(_PREFIX)

# these are magics that can modify the dependency structure beacuse they
# may declare new variables or use existing ones as inputs
HAS_INLINE_PYTHON = {'%%capture', '%%timeit', '%%time', '%time', '%timeit'}


def comment_magics(nb):
    """
    Iterates over cells, commenting the ones with magics
    """
    nb = copy.deepcopy(nb)

    for cell in nb.cells:
        if cell.cell_type == 'code':
            cell['source'] = _comment_if_ipython_magic(cell['source'])

    return nb


def uncomment_magics(nb):
    """
    Iterates over cells, uncommenting the ones with magics
    """
    nb = copy.deepcopy(nb)

    for cell in nb.cells:
        if cell.cell_type == 'code':
            cell['source'] = _uncomment_magics_cell(cell['source'])

    return nb


def _delete_magic(line):
    """Returns an empty line if it starts with the # [magic] prefix
    """
    return '' if line.startswith(_PREFIX) else line


def _delete_magics_cell(source):
    """Reverts the comments applied to magics (cell level)
    """
    if not source:
        return source

    lines = source.splitlines()
    lines_new = [_delete_magic(line) for line in lines]
    return '\n'.join(lines_new).strip()


def _uncomment_magic(line):
    """Reverts the comments applied to magics (line level)
    """
    if line.startswith(_PREFIX):
        return line[_PREFIX_LEN:]

    parts = _is_commented_line_magic(line)

    if parts:
        code, magic = parts
        return f'{magic} {code}'
    else:
        return line


def _uncomment_magics_cell(source):
    """Reverts the comments applied to magics (cell level)
    """
    lines = source.splitlines()
    lines_new = [_uncomment_magic(line) for line in lines]
    return '\n'.join(lines_new)


def _comment(line):
    """Adds the # [magic] prefix (line level)
    """
    return f'# [magic] {line}'


def _comment_ipython_line_magic(line, magic):
    """Adds suffix to line magics: # [magic] %{name}

    Converts:
    %timeit x = 1

    Into:
    x = 1 # [magic] %timeit
    """
    return line.replace(magic, '').strip() + f' # [magic] {magic.strip()}'


def _comment_if_ipython_magic(source):
    """Comments lines into comments if they're IPython magics (cell level)
    """
    # TODO: support for nested cell magics. e.g.,
    # %%timeit
    # %%timeit
    # something()
    lines_out = []
    comment_rest = False

    # TODO: inline magics should add a comment at the end of the line, because
    # the python code may change the dependency structure. e.g.,
    # %timeit z = x + y -> z = x + y # [magic] %timeit
    # note that this only applies to inline magics that take Python code as arg

    # NOTE: magics can take inputs but their outputs ARE NOT saved. e.g.,
    # %timeit x = y + 1
    # running such magic requires having y but after running it, x IS NOT
    # declared. But this is magic dependent %time x = y + 1 will add x to the
    # scope

    for line in source.splitlines():
        cell_magic = _is_ipython_cell_magic(line)

        if comment_rest:
            lines_out.append(_comment(line))
        else:
            line_magic = _is_ipython_line_magic(line)

            # if line magic, comment line
            if line_magic:
                # NOTE: switched _comment_ipython_line_magic(line, line_magic)
                # for _comment
                lines_out.append(_comment(line))

            # if inline shell, comment line
            elif _is_inline_shell(line):
                lines_out.append(_comment(line))

            # if cell magic, comment line
            elif cell_magic in HAS_INLINE_PYTHON:
                lines_out.append(_comment(line))

            # if cell magic whose content *is not* Pytho, comment line and
            # all the remaining lines in the cell
            elif cell_magic:
                lines_out.append(_comment(line))
                comment_rest = True

            # otherwise, don't do anything
            else:
                lines_out.append(line)

    return '\n'.join(lines_out)


# NOTE: the code in the following lines is based on Ploomber's source code.
# We did that instead of importing it since Ploomber has many dependencies. At
# some point, we may integrate this into Ploomber and remove this.


def _is_commented_line_magic(source):
    """Determines if the source is an IPython cell magic. e.g.,

    %cd some-directory
    """
    m = re.match(_IS_COMMENTED_LINE_MAGIC, source)

    if not m:
        return False

    return m.group(1), m.group(2)


def _is_ipython_cell_magic(source):
    """Determines if the source is an IPython cell magic. e.g.,

    %cd some-directory
    """
    m = re.match(_IS_IPYTHON_CELL_MAGIC, source.lstrip())

    if not m:
        return False

    return m.group()


def _is_ipython_line_magic(line):
    """
    Determines if the source line is an IPython magic. e.g.,

    %%bash
    for i in 1 2 3; do
      echo $i
    done
    """
    m = re.match(_IS_IPYTHON_LINE_MAGIC, line)

    if not m:
        return False

    return m.group()


def _is_inline_shell(line):
    m = re.match(_IS_INLINE_SHELL, line)

    if not m:
        return False

    return m.group()



"""
Module to determine inputs and outputs from code snippets.
"""
from functools import reduce

import parso

from soorgeon import detect, definitions

_BUILTIN = set(__builtins__)


class ImportsParser:
    """Parses import statements to know which ones to inject to any given task

    Parameters
    ----------
    code_nb : str
        Notebook's source code
    """

    def __init__(self, code_nb):
        self._tree = parso.parse(code_nb)
        # maps defined names (from imports) to the source code
        self._name2code = definitions.from_imports(self._tree)

    def get_imports_cell_for_task(self, code_task):
        """
        Get the source code with the appropriate import statements for a task
        with the given code to work.
        """
        # NOTE: this was taken from jupyblog's source
        leaf = parso.parse(code_task).get_first_leaf()

        names = []

        while leaf:
            if leaf.type == 'name':
                names.append(leaf.value)

            leaf = leaf.get_next_leaf()

        imports = self._name2code

        # iterate over names defined by the imports and get the import
        # statement if content_subset uses it
        imports_to_use = []

        for name, import_code in imports.items():
            if name in names:
                imports_to_use.append(import_code)

        # remove duplicated elements but keep order, then join
        if imports:
            imports_to_use = ('\n'.join(list(dict.fromkeys(imports_to_use))) +
                              '\n\n\n').strip() or None
        else:
            imports_to_use = None

        # FIXME: once we parse the imports, we should remove them from the code
        # otherwise there are going to be duplicates

        return imports_to_use


def get_local_scope(leaf):
    """
    Returns a set of variables that are defined locally and thus should
    not be considered inputs (e.g., variables defined in a for loop)
    """
    parent = leaf.parent

    while parent:
        if parent.type == 'for_stmt':
            # call recursively for nested for loops to work
            return (find_for_loop_def_and_io(parent)[0]
                    | get_local_scope(parent.parent))

        # FIXME: this wont work with nested functions
        elif parent.type == 'funcdef':
            def_names = [
                c.get_defined_names() for c in parent.children[2].children
                if c.type == 'param'
            ]

            flatten = [name.value for sub in def_names for name in sub]

            return set(flatten)

        parent = parent.parent

    return set()


def find_for_loop_def_and_io(for_stmt, local_scope=None):
    """
    Return a set with the definitions and inputs a for loop. e.g.,
    for x, (y, z) in something() returns {'x', 'y', 'z'}, set()
    for i in range(input_) returns {'i'}, {'input_'}
    """
    # TODO: add a only_input flag for cases where we dont care about
    # parsin outputs
    if for_stmt.type != 'for_stmt':
        raise ValueError(f'Expected a node with type "for_stmt", '
                         f'got: {for_stmt} with type {for_stmt.type}')

    local_scope = local_scope or set()

    # get the parts that we need
    _, node_definition, _, node_iterator, _, body_node = for_stmt.children

    defined = find_inputs(node_definition, parse_list_comprehension=False)
    iterator_in = find_inputs(node_iterator, parse_list_comprehension=False)

    body_in, body_out = find_inputs_and_outputs_from_leaf(
        body_node.get_first_leaf(),
        local_scope=defined,
        leaf_end=body_node.get_last_leaf())

    # Strictly speaking variables defined after the for keyword are also
    # outputs, since they're available after the loop ends (with the loop's
    # last value, however, we don't consider them here)
    return defined, (iterator_in | body_in) - local_scope, body_out


def _find_type_value_idx_in_children(type_, value, node):
    for idx, child in enumerate(node.children):
        if (child.type, getattr(child, 'value', None)) == (type_, value):
            return idx

    return None


def _process_context(context):
    if ('keyword', 'as') in ((n.type, getattr(n, 'value', None))
                             for n in context.children):
        node_expression, _, node_definition = context.children
        defined = find_inputs(node_definition, parse_list_comprehension=False)
    else:
        node_expression = context
        defined = set()

    # FIXME: this could have a list comprehension
    exp = find_inputs(node_expression, parse_list_comprehension=False)

    return exp, defined


def find_f_string_inputs(fstring_start, local_scope=None):
    if fstring_start.type != 'fstring_start':
        raise ValueError(
            f'Expected a node with type "fstring_start", '
            f'got: {fstring_start} with type {fstring_start.type}')

    f_string = fstring_start.parent

    names = find_inputs(f_string, parse_list_comprehension=False)

    return names - local_scope


def find_context_manager_def_and_io(with_stmt, local_scope=None):
    if with_stmt.type != 'with_stmt':
        raise ValueError(f'Expected a node with type "with_stmt", '
                         f'got: {with_stmt} with type {with_stmt.type}')

    local_scope = local_scope or set()

    idx_colon = _find_type_value_idx_in_children('operator', ':', with_stmt)

    # get children that are relevant (ignore with keyword, commads, and colon
    # operator)
    contexts = with_stmt.children[1:idx_colon:2]

    body_node = with_stmt.children[-1]

    exp, defined = set(), set()

    for context in contexts:
        exp_, defined_ = _process_context(context)
        exp = exp | exp_
        defined = defined | defined_

    body_in, body_out = find_inputs_and_outputs_from_leaf(
        body_node.get_first_leaf(),
        local_scope=defined,
        leaf_end=body_node.get_last_leaf())

    return defined, (exp | body_in) - local_scope, body_out


def find_function_scope_and_io(funcdef, local_scope=None):
    """Find function scope, input, and output

    Returns
    -------
    set
        Parameters in the signature

    set
        Variable from the global scope (used in the body, but missing in the
        signature), also includes return annotations (e.f., fn() -> str)

    set
        Variables declared in the body
    """
    if funcdef.type != 'funcdef':
        raise ValueError(f'Expected a node with type "funcdef", '
                         f'got: {funcdef} with type {funcdef.type}')

    local_scope = local_scope or set()

    # regular definition
    if len(funcdef.children) == 5:
        _, _, node_signature, _, body_node = funcdef.children
        annotation_return = None

    # return annotation
    elif len(funcdef.children) == 7:
        node_signature = funcdef.children[2]
        annotation_return = funcdef.children[4]
        body_node = funcdef.children[6]

    # FIXME: test what happens if they user has a list comprehension as
    # argument. e.g. fn(x=[1,2,3])
    parameters = find_inputs(node_signature,
                             parse_list_comprehension=False,
                             allow_kwargs=True)

    body_in, body_out = find_inputs_and_outputs_from_leaf(
        body_node.get_first_leaf(),
        local_scope=parameters,
        leaf_end=body_node.get_last_leaf())

    if annotation_return:
        body_in = body_in | find_inputs(annotation_return,
                                        parse_list_comprehension=False)

    return parameters, body_in - local_scope, body_out


# TODO: add unit tests
def find_lambda_scope_and_inputs(lambda_, local_scope=None):
    if lambda_.type != 'lambdef':
        raise ValueError(f'Expected a lambdef, got {lambda_}')

    local_scope = local_scope or set()

    _, node_signature, _, body_node = lambda_.children

    # FIXME: test what happens if they user has a list comprehension as
    # argument. e.g. fn(x=[1,2,3])
    parameters = find_inputs(node_signature,
                             parse_list_comprehension=False,
                             allow_kwargs=True)

    body_in = find_inputs(body_node)

    return parameters, body_in - local_scope - parameters


def _flatten_sync_comp_for(node):
    if node.type != 'sync_comp_for':
        raise ValueError('Expected node type to be '
                         f'"syncompfor" but got: {node.type}')

    total = [node]

    for child in node.children:
        if child.type == 'sync_comp_for':
            nodes = _flatten_sync_comp_for(child)
            total += nodes

    return total


def _find_sync_comp_for_inputs_and_scope(synccompfor):
    """
    Find inputs and scope for a sync comp for node, note that this only
    parses a single node, for parsing nested ones use find_comprehension_inputs
    """
    # these are the variables that the list comprehension declares
    declared = find_inputs(synccompfor.children[1],
                           parse_list_comprehension=False)

    # parse the variables in the right expression
    # e,g, given: [x for x in expression(10)]
    # the expression is expression(10)
    # the expression part should be the element at index 3, note that this
    # is not the same as getting the last one because if the list comprehension
    # has an 'if' statement, that will be the last element
    inputs_right = find_inputs(synccompfor.children[3],
                               parse_list_comprehension=False)

    return inputs_right, declared


def find_comprehension_inputs(node):
    """Find inpust for list/set/dict comprehension or generator
    """
    types = {'testlist_comp', 'dictorsetmaker'}

    if node.type not in types:
        raise ValueError('Expected node type be one of '
                         f'{types!r}, got: {node.type}')

    # list/set comprehension or generator
    if len(node.children) == 2:
        compfor, synccompfor = node.children

        # parse the variables in the left expression
        # e.g., [expression(x) for x in range(10)]

        try:
            inputs_left = find_inputs(compfor, parse_list_comprehension=False)
        except AttributeError:
            # if there isn't an expression but a single variable, we just
            # get the name e.g., [x for x in range(10)]
            inputs_left = {compfor.value}

    # dict comprehension contains:
    # key: value
    else:
        synccompfor = node.children[-1]

        inputs_left = find_inputs_for_each(node.children[:-1],
                                           parse_list_comprehension=False)

    inputs, declared = set(), set()

    for node in _flatten_sync_comp_for(synccompfor):
        inputs_, declared_ = _find_sync_comp_for_inputs_and_scope(node)
        inputs = inputs | inputs_
        declared = declared | declared_

    return (inputs_left | inputs) - declared


def find_inputs_for_each(nodes,
                         parse_list_comprehension=True,
                         only_getitem_and_attribute_access=False):
    """Like find_inputs, but takes a list of nodes
    """
    inputs = set()

    for node in nodes:
        inputs_new = find_inputs(
            node,
            parse_list_comprehension=parse_list_comprehension,
            only_getitem_and_attribute_access=only_getitem_and_attribute_access
        )
        inputs = inputs | inputs_new

    return inputs


def find_inputs(node,
                parse_list_comprehension=True,
                only_getitem_and_attribute_access=False,
                allow_kwargs=False):
    """
    Extract inputs from an expression
    e.g. function(x, y) returns {'function', 'x', 'y'}

    Parameters
    ----------
    parse_list_comprehension : bool, default=True
        Whether to parse any list comprehension if node is inside one
    """

    names = []

    leaf = node.get_first_leaf()
    # stop when you reach the end of the expression
    last = node.get_last_leaf()

    while leaf:
        if detect.is_comprehension(leaf):
            inputs = find_comprehension_inputs(leaf.get_next_sibling())
            names.extend(list(inputs))
            leaf = leaf.parent.get_last_leaf()
        if detect.is_lambda(leaf):
            scope, inputs = find_lambda_scope_and_inputs(leaf.parent)
            names.extend(list(inputs - scope))
            # lambda's last leaf is the next one after the last in the
            # lambda node
            leaf = leaf.parent.get_last_leaf().get_next_leaf()
        # something else
        else:
            # ignore f-string format specs {number:.2f}
            # and f-string conversions {object!r}
            if leaf.parent.type in {
                    'fstring_format_spec', 'fstring_conversion'
            }:
                leaf = leaf.get_next_leaf()
                continue

            # is this a kwarg?
            try:
                key_arg = leaf.get_next_leaf().value == '='
            except Exception:
                key_arg = False

            if allow_kwargs:
                key_arg = False

            # is this an attribute?
            try:
                is_attr = leaf.get_previous_leaf().value == '.'
            except Exception:
                is_attr = False

            if (leaf.type == 'name' and not key_arg and not is_attr
                    and leaf.value not in _BUILTIN):
                # not allowing reads, check that this is not geitem
                # or that is accessing an attribute in the next leaf
                try:
                    is_getitem = leaf.get_next_leaf().value == '['
                except Exception:
                    is_getitem = False

                try:
                    is_accessing_attr = leaf.get_next_leaf().value == '.'
                except Exception:
                    is_accessing_attr = False

                if (only_getitem_and_attribute_access
                        and (is_getitem or is_accessing_attr)):
                    names.append(leaf.value)
                elif not only_getitem_and_attribute_access:
                    names.append(leaf.value)

            if leaf is last:
                break

            leaf = leaf.get_next_leaf()

    return set(names)


def find_inputs_and_outputs(code_str, local_scope=None):
    """
    Given a Python code string, find which variables the code consumes (not
    declared in the snipped) and which ones it exposes (declared in the
    snippet)

    Parameters
    ----------
    local_scope : set
        Names that should not be considered inputs
    """
    tree = parso.parse(code_str)
    return find_inputs_and_outputs_from_tree(tree, local_scope=local_scope)


def find_inputs_and_outputs_from_tree(tree, local_scope=None):
    leaf = tree.get_first_leaf()
    # NOTE: we use this in find_inputs_and_outputs and ImportParser, maybe
    # move the functionality to a class so we only compute it once
    defined_names = set(definitions.from_imports(tree)) | set(
        definitions.from_def_and_class(tree))

    local_scope = local_scope or set()

    return find_inputs_and_outputs_from_leaf(leaf,
                                             local_scope=local_scope
                                             | defined_names)


# FIXME: try nested functions, and also functions inside for loops and loops
# inside functions
def find_inputs_and_outputs_from_leaf(leaf, local_scope=None, leaf_end=None):
    """
    Find inputs and outputs. Starts parsing at the given leaf
    """
    local_scope = local_scope or set()

    inputs, outputs = [], set()

    local_variables = set()

    def clean_up_candidates(candidates, *others):
        candidates = candidates - set(_BUILTIN)
        candidates = candidates - set(local_scope)
        candidates = candidates - outputs

        for another in others:
            candidates = candidates - another

        return candidates

    while leaf:
        _inside_funcdef = detect.is_inside_funcdef(leaf)

        if not _inside_funcdef:
            local_variables = set()

        if detect.is_f_string(leaf):
            candidates_in = find_f_string_inputs(leaf, local_scope=local_scope)
            inputs.extend(clean_up_candidates(candidates_in, local_variables))
            # jump to the end of the f-string
            leaf = leaf.parent.get_last_leaf()
        elif detect.is_lambda(leaf):
            # FIXME: i think is hould also pass the current foudn inputs
            # to local scope - write a test to break this
            _, candidates_in = find_lambda_scope_and_inputs(
                leaf.parent, local_scope=local_scope)
            inputs.extend(clean_up_candidates(candidates_in, local_variables))
            # lambda's last leaf is the next one after the last in the
            # lambda node
            leaf = leaf.parent.get_last_leaf().get_next_leaf()
        elif detect.is_for_loop(leaf):
            # FIXME: i think is hould also pass the current foudn inputs
            # to local scope - write a test to break this
            (_, candidates_in, candidates_out) = find_for_loop_def_and_io(
                leaf.parent, local_scope=local_scope)
            inputs.extend(clean_up_candidates(candidates_in, local_variables))
            outputs = outputs | candidates_out
            # jump to the end of the foor loop
            leaf = leaf.parent.get_last_leaf()
        elif detect.is_context_manager(leaf):
            # FIXME: i think is hould also pass the current foudn inputs
            # to local scope - write a test to break this
            (_, candidates_in,
             candidates_out) = find_context_manager_def_and_io(
                 leaf.parent, local_scope=local_scope)
            inputs.extend(clean_up_candidates(candidates_in, local_variables))
            outputs = outputs | candidates_out
            # jump to the end of the foor loop
            leaf = leaf.parent.get_last_leaf()
        elif detect.is_funcdef(leaf):
            # NOTE: we're verifying that the function does not use any global
            # variables so candidates_in should only contain the return
            # annotations
            # FIXME: i think is hould also pass the current foudn inputs
            # to local scope - write a test to break this
            (_, candidates_in, candidates_out) = find_function_scope_and_io(
                leaf.parent, local_scope=local_scope)
            inputs.extend(clean_up_candidates(candidates_in, local_variables))
            outputs = outputs | candidates_out
            # jump to the end of the function definition loop
            leaf = leaf.parent.get_last_leaf()
        elif detect.is_classdef(leaf):
            # TODO: parse annotations

            leaf = leaf.parent.get_last_leaf()
        elif detect.is_comprehension(leaf):
            inputs_new = find_comprehension_inputs(leaf.get_next_sibling())
            inputs.extend(clean_up_candidates(inputs_new, local_variables))
            leaf = leaf.parent.get_last_leaf()

        # the = operator is an indicator of [outputs] = [inputs]
        elif leaf.type == 'operator' and leaf.value == '=':
            next_s = leaf.get_next_sibling()
            previous = leaf.get_previous_leaf()

            # Process inputs
            inputs_current = find_inputs(next_s)
            inputs_current = inputs_current - set(_BUILTIN)
            inputs_current = inputs_current - set(local_scope)

            for variable in inputs_current:
                # check if we're inside a for loop and ignore variables
                # defined there

                # only mark a variable as input if it hasn't been defined
                # locally
                if (variable not in outputs and variable not in local_scope
                        and variable not in local_variables):
                    inputs.append(variable)

            # Process outputs

            # ignore keyword arguments, they aren't outputs
            # e.g. 'key' in something(key=value)
            # also ignore previous if modifying a variable defined in the
            # current section
            # e.g.,
            # a = {}
            # a['x'] = 1
            # a.b = 1

            if previous.parent.type != 'argument':

                prev_sibling = leaf.get_previous_sibling()

                # if we're inside a function, we're not storing outputs
                # but local variables
                target = local_variables if _inside_funcdef else outputs

                # existing variables (definde in the current
                # section) as outputs
                modified = _get_modified_objects(leaf, outputs, local_scope)

                # check if assigning multiple values
                # e.g., a, b = 1, 2 (testlist_star_expr)
                # [a, b] = 1, 2 (atom)
                # (a, b) = 1, 2 (atom)
                if prev_sibling.type in {'testlist_star_expr', 'atom'}:
                    target = target | set(
                        name.value
                        for name in prev_sibling.parent.get_defined_names())
                # nope, only one value
                elif prev_sibling.type == 'atom_expr':
                    target = target | (find_inputs(
                        prev_sibling, parse_list_comprehension=False) -
                                       modified)
                elif previous.value not in modified:
                    target.add(previous.value)

                if _inside_funcdef:
                    local_variables = target
                else:
                    # before modifying the current outputs, check if there's
                    # anything on the left side of the = token that it's
                    # mutating a variable e.g.,:
                    # object.attribute = value
                    # or
                    # object['key'] = value
                    # this is the only case where something on the left side
                    # can be considered an input but only if the object hasn't
                    # been declared so far
                    inputs_candidates = find_inputs(
                        prev_sibling,
                        parse_list_comprehension=False,
                        only_getitem_and_attribute_access=True)

                    # add to inputs if they haven't been locally defined or
                    # modified
                    inputs_new = inputs_candidates - outputs - modified
                    inputs.extend(inputs_new)

                    outputs = target

        # Process inputs scenario #2 - there is not '=' token but a function
        # call. variables inside function calls are inputs
        # e.g., some_function(df)
        # in this case, df is considered an input, except if it has been
        # locally defined.
        # e.g.,
        # df = something()
        # some_function(df)
        # FIXME: this is redundant because when we identify the '=' token, we
        # go to the first conditional, and the next leaf is the function call
        # so then we go into this conditional - we're skipping the left part
        # but not the right part of = yet
        elif (leaf.type == 'name' and (detect.is_inside_function_call(leaf)
                                       or detect.is_accessing_variable(leaf)
                                       or detect.is_inside_funcdef(leaf))
              # skip if this is to the left of an '=', because we'll check it
              # when we get to that token since it'll go to the first
              # conditional
              and not detect.is_left_side_of_assignment(leaf) and
              not detect.is_inside_list_comprehension(leaf) and
              leaf.value not in outputs and leaf.value not in local_scope and
              leaf.value not in _BUILTIN and leaf.value not in local_scope and
              leaf.value not in local_variables):
            inputs.extend(find_inputs(leaf))

        if leaf_end and leaf == leaf_end:
            break

        leaf = leaf.get_next_leaf()

    return set(inputs), outputs


def _get_modified_objects(leaf, outputs, names_from_imports):
    names = []
    sibling = leaf.get_previous_sibling()
    current = sibling.get_first_leaf()
    leaf_last = sibling.get_last_leaf()
    existing = outputs | names_from_imports

    # iterate over leaves and grab names since the assignment may be modifying
    # more than one object
    while current:
        if current.type == 'name':
            if current.value in existing:
                names.append(current.value)

        if current == leaf_last:
            break

        current = current.get_next_leaf()

    return set(names)


def _map_outputs(name, outputs):
    return [(out, name) for out in outputs]


def _get_upstream(name, inputs, providers):
    return [providers.get(input_, name) for input_ in inputs]


class ProviderMapping:
    """
    Determines which task produces a given variable to establish the upstream
    relationship

    Parameters
    ----------
    io : dict
        {task_name: (inputs, outputs), ...} mapping. Note that order is
        important, and is assumed that the order of the task_name keys matches
        the order of appearance of their corresponding cells in the notebook.
    """

    def __init__(self, io):
        self._io = io

    def _providers_for_task(self, name):
        """
        Returns a subset of io, only considering tasks that appear earlier
        in the notebook
        """
        out = {}

        for key, value in self._io.items():
            if key == name:
                break

            out[key] = value

        return _find_providers(out)

    def get(self, variable, task_name):
        """
        Return the provider of a certain variable for a task with name
        task_name. The function depends on task_name because if two tasks
        expose a variable with the same name, we need to resolve to the
        one closest to the task_name, by considering all previous sections
        in the notebook
        """
        providers = self._providers_for_task(task_name)

        provider = providers.get(variable)

        if not provider:
            raise KeyError(f'Error parsing inputs for section {task_name!r} '
                           'notebook: could not find an earlier section '
                           f'declaring variable {variable!r}')

        return provider


class DefinitionsMapping:
    """
    Returns the available names (from import statements, function and class
    definitions) available for a given
    snippet. We use this to determine which names should not be considered
    inputs in tasks, since they are module names
    """

    def __init__(self, snippets):
        self._names = {
            name: set(definitions.find_defined_names(parso.parse(code)))
            for name, code in snippets.items()
        }

    def get(self, name):
        out = set()

        for key, value in self._names.items():
            if key == name:
                break

            out = out | value

        return out


def find_upstream(snippets):
    """
    Parameters
    ----------
    snippets : dict
        {snippet_name: snippet, ...}
    """
    io = find_io(snippets)

    providers = ProviderMapping(io)

    upstream = {
        snippet_name: list(set(_get_upstream(snippet_name, v[0], providers)))
        for snippet_name, v in io.items()
    }

    return upstream


def _find_providers(io):
    # variable -> snippet that defines variable mapping
    providers = [
        _map_outputs(snippet_name, v[1]) for snippet_name, v in io.items()
    ]

    providers = dict([i for sub in providers for i in sub])

    return providers


def find_io(snippets):
    """
    Generates a {snippet_name: (inputs, outputs), ...} mapping where inputs
    are the variables that snippet_name requires to work and outputs the ones
    that it creates
    """
    im = DefinitionsMapping(snippets)

    # FIXME: find_upstream already calls this, we should only compute it once
    io = {
        snippet_name: find_inputs_and_outputs(snippet,
                                              local_scope=im.get(snippet_name))
        for snippet_name, snippet in snippets.items()
    }

    return io


def prune_io(io):
    """
    Prunes an io mapping (as generated by find_io) that removes outputs that
    aren't used by any downstream task
    """
    # FIXME: order is important. if snippet at index i exports a, b, c, we
    # should only consider snippets beginning at index i + 1

    used = reduce(lambda x, y: x | y, (inputs for inputs, _ in io.values()))

    pruned = {}

    for key, (inputs, outputs) in io.items():
        outputs_pruned = outputs & used
        pruned[key] = (inputs, outputs_pruned)

    return pruned


def _leaf_iterator(tree):
    leaf = tree.get_first_leaf()

    while leaf:
        yield leaf

        leaf = leaf.get_next_leaf()


def remove_imports(code_str):
    """
    Remove all import statements from a code string
    """
    tree = parso.parse(code_str)

    to_remove = []

    for leaf in _leaf_iterator(tree):
        if leaf.parent.type in {'import_name', 'import_from'}:
            to_remove.append(leaf)

    for leaf in to_remove:
        leaf.parent.children = []

    return tree.get_code()



"""
We want to parse a series of Python snippets (one per notebook section -
splitted by H2 headers). For each snippet, we need to know which variables
they need (from previous sections) and which ones they expose (available for
upcoming sections). There are a few prime cases we want to cover.

1. Assignment: a = 1
2. Function calling: function(1, b=2)
3. If/While statements
4. List comprehensions
5. For loops
6. Function definitions

The first three are simple, however list comprehensions, for loops and
functions introduce a new complexity: they can be nested. Example:

for i in range(10):
    for j in range(10):
        print(i, j)

Hence, we need to keep into account that they define a local scope, hence
some of the variables should not be considered exposed to upcoming tasks.
For example, in our nested loop i and j are not considered exposed.
if/while statements can also be arbitrarily nested but they do not define a
local scope, so we don't have that complication.

To deal with this possibly nested logic, we have a function
(find_inputs_and_outputs_from_tree) that goes through every leaf in the ast,
and determines which sub function to execute. For example, if it finds a 'for'
keyword, it calls find_for_loop_def_and_io, which parses a for
structure.

Note that find_inputs_and_outputs_from_tree and
find_for_loop_def_and_io may call find_inputs_and_outputs_from_tree
again, because such a function deals with arbitrary structures. However,
when doing nested calls to find_inputs_and_outputs_from_tree, one should
set the leaf parameter to tell the function when to stop and return.

The second type of functions are detectors, which determine which
sub-function to call (e.g., is_for_loop identifies if we're about to
enter a for loop). These detectors should return False if we're not inside
the structure they identify, if we're in a given structure, they should return
the last leaf, so we can pass this to the parsing function in the until leaf
argument.
"""

__version__ = '0.0.15'



def first_expr_stmt_parent(node):
    parent = node.parent

    if not parent:
        return None

    while parent.type != 'expr_stmt':
        parent = parent.parent

        if not parent:
            break

    return parent


def first_non_atom_expr_parent(node):
    parent = node.parent

    # e.g., [x.attribute for x in range(10)]
    # x.attribute is an atom_expr
    while parent.type == 'atom_expr':
        parent = parent.parent

    return parent



from functools import reduce
from isort import place_module


# NOTE: we use this in find_inputs_and_outputs and ImportParser, maybe
# move the functionality to a class so we only compute it once
def from_imports(tree):
    # build a defined-name -> import-statement-code mapping. Note that
    # the same code may appear more than once if it defines more than one name
    # e.g. from package import a, b, c
    imports = [{
        name.value: import_.get_code().rstrip()
        for name in import_.get_defined_names()
    } for import_ in tree.iter_imports()]

    if imports:
        imports = reduce(lambda x, y: {**x, **y}, imports)
    else:
        imports = {}

    return imports


def packages_used(tree):
    """
    Return a list of the packages used, correcting for some packages whose
    module name does not match the PyPI package (e.g., sklearn -> scikit-learn)

    Returns None if fails to parse them
    """
    pkg_name = {
        'sklearn': 'scikit-learn',
    }

    def flatten(elements):
        return [i for sub in elements for i in sub]

    def extract_names(import_):
        if import_.type == 'name':
            return [import_.value]
        elif import_.type in {'dotted_name', 'dotted_as_name'}:
            return [import_.children[0].value]

        second = import_.children[1]

        if second.type in {'dotted_name', 'dotted_as_name'}:
            return extract_names(second.children[0])
        elif second.type == 'dotted_as_names':
            # import a as something, b as another
            return flatten([
                extract_names(node.children[0])
                for i, node in enumerate(second.children) if i % 2 == 0
            ])
        else:
            return [second.value]

    pkgs = flatten([extract_names(import_) for import_ in tree.iter_imports()])

    # replace using pkg_name mapping and ignore standard lib
    pkgs_final = [
        pkg_name.get(name, name) for name in pkgs
        if place_module(name) == 'THIRDPARTY'
    ]

    # remove duplicates and sort
    return sorted(set(pkgs_final))


def from_def_and_class(tree):
    fns = {
        fn.name.value: fn.get_code().rstrip()
        for fn in tree.iter_funcdefs()
    }

    classes = {
        class_.name.value: class_.get_code().rstrip()
        for class_ in tree.iter_classdefs()
    }

    return {**fns, **classes}


def find_defined_names(tree):
    return {**from_imports(tree), **from_def_and_class(tree)}



"""
NOTE: this was taken from Ploomber, we copied it instead of importing it
since ploomber has many dependencies, at some point we'll either move
this functionality into soorgeon or create a ploomber-core package
and move things over there
"""
import warnings
from io import StringIO

from pyflakes import api as pyflakes_api
from pyflakes.reporter import Reporter
from pyflakes.messages import (UndefinedName, UndefinedLocal,
                               DuplicateArgument, ReturnOutsideFunction,
                               YieldOutsideFunction, ContinueOutsideLoop,
                               BreakOutsideLoop)

from soorgeon.exceptions import InputWontRunError, InputSyntaxError

# messages: https://github.com/PyCQA/pyflakes/blob/master/pyflakes/messages.py
_ERRORS = (
    UndefinedName,
    UndefinedLocal,
    DuplicateArgument,
    ReturnOutsideFunction,
    YieldOutsideFunction,
    ContinueOutsideLoop,
    BreakOutsideLoop,
)


def _process_messages(mesages):
    return '\n'.join(str(msg) for msg in mesages)


def process_errors_and_warnings(messages):
    errors, warnings = [], []

    for message in messages:
        if isinstance(message, _ERRORS):
            errors.append(message)
        else:
            warnings.append(message)

    return _process_messages(errors), _process_messages(warnings)


# https://github.com/PyCQA/pyflakes/blob/master/pyflakes/reporter.py
class MyReporter(Reporter):

    def __init__(self):
        self._stdout = StringIO()
        self._stderr = StringIO()
        self._stdout_raw = []
        self._unexpected = False
        self._syntax = False

    def flake(self, message):
        self._stdout_raw.append(message)
        self._stdout.write(str(message))
        self._stdout.write('\n')

    def unexpectedError(self, *args, **kwargs):
        """pyflakes calls this when ast.parse raises an unexpected error
        """
        self._unexpected = True
        return super().unexpectedError(*args, **kwargs)

    def syntaxError(self, *args, **kwargs):
        """pyflakes calls this when ast.parse raises a SyntaxError
        """
        self._syntax = True
        return super().syntaxError(*args, **kwargs)

    def _seek_zero(self):
        self._stdout.seek(0)
        self._stderr.seek(0)

    def _make_error_message(self, error):
        return ('Errors detected in your source code:'
                f'\n{error}\n\n'
                '(ensure that your notebook executes from top-to-bottom '
                'and try again)')

    def _check(self):
        self._seek_zero()

        # syntax errors are stored in _stderr
        # https://github.com/PyCQA/pyflakes/blob/master/pyflakes/api.py

        error_message = '\n'.join(self._stderr.readlines())

        if self._syntax:
            raise InputSyntaxError(self._make_error_message(error_message))
        elif self._unexpected:
            warnings.warn('An unexpected error happened '
                          f'when analyzing code: {error_message.strip()!r}')
        else:
            errors, warnings_ = process_errors_and_warnings(self._stdout_raw)

            if warnings_:
                warnings.warn(warnings_)

            if errors:
                raise InputWontRunError(self._make_error_message(errors))


def check_notebook(nb):
    """
    Run pyflakes on a notebook, wil catch errors such as missing passed
    parameters that do not have default values

    Parameters
    ----------
    nb : NotebookNode
        Notebook object. Must have a cell with the tag "parameters"

    filename : str
        Filename to identify pyflakes warnings and errors

    Raises
    ------
    SyntaxError
        If the notebook's code contains syntax errors

    TypeError
        If params and nb do not match (unexpected or missing parameters)

    RenderError
        When certain pyflakes errors are detected (e.g., undefined name)
    """
    # concatenate all cell's source code in a single string
    source_code = '\n'.join(c['source'] for c in nb.cells
                            if c.cell_type == 'code')

    # this objects are needed to capture pyflakes output
    reporter = MyReporter()

    # run pyflakes.api.check on the source code
    pyflakes_api.check(source_code, filename='', reporter=reporter)

    reporter._check()
